"""Sports reference player model."""

# pylint: disable=too-many-arguments,unused-argument,line-too-long,duplicate-code
import datetime
import http
import logging
from urllib.parse import unquote

import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup

from ...cache import MEMORY
from ...session import DEFAULT_TIMEOUT
from ..player_model import PlayerModel

_FIX_URLS = {
    "https://www.sports-reference.com/cbb/players/leyla-öztürk-1.html": "https://www.sports-reference.com/cbb/players/leyla-ozturk-1.html",
    "https://www.sports-reference.com/cbb/players/vianè-cumber-1.html": "https://www.sports-reference.com/cbb/players/viane-cumber-1.html",
    "https://www.sports-reference.com/cbb/players/cia-eklof-1.html": "https://www.sports-reference.com/cbb/players/cia-eklöf-1.html",
    "https://www.sports-reference.com/cbb/players/chae-harris-1.html": "https://www.sports-reference.com/cbb/players/cha%C3%A9-harris-1.html",
    "https://www.sports-reference.com/cbb/players/tilda-sjokvist-1.html": "https://www.sports-reference.com/cbb/players/tilda-sjökvist-1.html",
    "https://www.sports-reference.com/cbb/players/hana-muhl-1.html": "https://www.sports-reference.com/cbb/players/hana-mühl-1.html",
    "https://www.sports-reference.com/cbb/players/noa-comesaña-1.html": "https://www.sports-reference.com/cbb/players/noa-comesana-1.html",
    "https://www.sports-reference.com/cbb/players/nadège-jean-1.html": "https://www.sports-reference.com/cbb/players/nadege-jean-1.html",
}


def _fix_url(url: str) -> str:
    url = unquote(url)
    url = url.replace("é", "e")
    url = url.replace("ć", "c")
    url = url.replace("ã", "a")
    url = url.replace("á", "a")
    url = url.replace("á", "a")
    url = url.replace("ö", "o")
    url = url.replace("ü", "u")

    url = url.replace("Ã©", "é")
    url = url.replace("Ã¶", "ö")
    url = url.replace("Ã¼", "ü")

    return _FIX_URLS.get(url, url)


def _create_sportsreference_player_model(
    session: requests_cache.CachedSession,
    player_url: str,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> PlayerModel | None:
    """Create a player model from NCAAB sports reference."""
    player_url = _fix_url(player_url)
    response = session.get(player_url, timeout=DEFAULT_TIMEOUT)
    # Some players can't be accessed on sports reference
    if response.status_code == http.HTTPStatus.FORBIDDEN:
        logging.warning("Cannot access player at URL %s", player_url)
        return None
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")
    h1 = soup.find("h1")
    if h1 is None:
        logging.warning("h1 is null for %s", player_url)
        return None
    name = h1.get_text().strip()
    return PlayerModel(
        identifier=name,
        jersey=None,
        kicks=None,
        fumbles=None,
        fumbles_lost=None,
        field_goals=fg.get(name),
        field_goals_attempted=fga.get(name),
        offensive_rebounds=offensive_rebounds.get(name),
        assists=assists.get(name),
        turnovers=turnovers.get(name),
        name=name,
        marks=None,
        handballs=None,
        disposals=None,
        goals=None,
        behinds=None,
        hit_outs=None,
        tackles=None,
        rebounds=None,
        insides=None,
        clearances=None,
        clangers=None,
        free_kicks_for=None,
        free_kicks_against=None,
        brownlow_votes=None,
        contested_possessions=None,
        uncontested_possessions=None,
        contested_marks=None,
        marks_inside=None,
        one_percenters=None,
        bounces=None,
        goal_assists=None,
        percentage_played=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_player_model(
    session: requests_cache.CachedSession,
    player_url: str,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> PlayerModel | None:
    return _create_sportsreference_player_model(
        session, player_url, fg, fga, offensive_rebounds, assists, turnovers
    )


def create_sportsreference_player_model(
    session: requests_cache.CachedSession,
    player_url: str,
    dt: datetime.datetime,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> PlayerModel | None:
    """Create a player model from sports reference."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_player_model(
            session, player_url, fg, fga, offensive_rebounds, assists, turnovers
        )
    with session.cache_disabled():
        return _create_sportsreference_player_model(
            session, player_url, fg, fga, offensive_rebounds, assists, turnovers
        )



"""OddsPortal odds model."""

import datetime

from ...cache import MEMORY
from ..team_model import OddsModel
from .oddsportal_bookie_model import create_oddsportal_bookie_model


@MEMORY.cache
def create_oddsportal_odds_model(
    odds: float, dt: datetime.datetime, bookie_name: str, bookie_id: str
) -> OddsModel:
    """Create an odds model based off aus sports betting."""
    bookie = create_oddsportal_bookie_model(bookie_name, bookie_id)
    return OddsModel(odds=odds, bookie=bookie, dt=dt)  # pyright: ignore



"""Tests for the sportsreference player model class."""
import datetime
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.sportsreference.sportsreference_player_model import create_sportsreference_player_model


class TestSportsReferencePlayerModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_field_goals(self):
        url = "https://www.basketball-reference.com/players/b/barnesc01.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "barnesc01.html"), "rb") as f:
                m.get(url, content=f.read())
            player_model = create_sportsreference_player_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 0),
                {"Scottie Barnes": 8},
                {},
                {},
                {},
                {},
            )
            self.assertEqual(player_model.field_goals, 8)



"""The prototype class for news."""

import datetime

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType


class NewsModel(BaseModel):
    """The serialisable news class."""

    title: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    published: datetime.datetime
    summary: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    source: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})



"""Tests for the NBA NBA game model class."""
import datetime
import unittest

import requests_mock
import requests_cache
from sportsball.data.nba.nba.nba_nba_game_model import create_nba_nba_game_model
from sportsball.data.league import League
import pandas as pd


class TestNBANBAGameModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")

    def test_identifier(self):
        dt = datetime.datetime(2010, 10, 10, 10, 10, 00)
        with requests_mock.Mocker() as m:
            m.get("https://news.google.com/rss/search?q=%22The+A%27s%22+%2B+%28sport+OR+nba+OR+%22National+Basketball+League%22%29+after%3A2010-10-08+before%3A2010-10-09&ceid=US:en&hl=en&gl=US")
            m.get("https://news.google.com/rss/search?q=%22The+B%27s%22+%2B+%28sport+OR+nba+OR+%22National+Basketball+League%22%29+after%3A2010-10-08+before%3A2010-10-09&ceid=US:en&hl=en&gl=US")
            game_model = create_nba_nba_game_model(
                pd.Series(data=[
                    "22022",
                    str(dt.date()),
                    "a",
                    "The A's",
                    20.0,
                    3.0,
                    "b",
                    "The B's",
                    60.0,
                    "1",
                    4.0,
                    30.0,
                    30.0,
                    10.0,
                    12.0,
                    20.0,
                    30.0,
                    12,
                    15,
                ], index=[
                    "SEASON_ID",
                    "GAME_DATE",
                    "TEAM_ID_A",
                    "TEAM_NAME_A",
                    "PTS_A",
                    "FGM_A",
                    "TEAM_ID_B",
                    "TEAM_NAME_B",
                    "PTS_B",
                    "GAME_ID",
                    "FGM_B",
                    "FGA_A",
                    "FGA_B",
                    "OREB_A",
                    "OREB_B",
                    "AST_A",
                    "AST_B",
                    "TOV_A",
                    "TOV_B",
                ]),
                League.NBA,
                0,
                1,
                self.session,
                dt,
                "00",
            )
            self.assertEqual(game_model.dt.date(), dt.date())



"""ESPN game model."""

# pylint: disable=too-many-arguments
import datetime
from typing import Any, Dict

import pytest_is_running
import requests_cache
from dateutil.parser import parse

from ...cache import MEMORY
from ..game_model import GameModel, localize
from ..league import League
from ..odds_model import OddsModel
from ..season_type import SeasonType
from ..team_model import TeamModel
from ..venue_model import VenueModel
from .espn_bookie_model import create_espn_bookie_model
from .espn_odds_model import MONEYLINE_KEY, create_espn_odds_model
from .espn_team_model import create_espn_team_model
from .espn_venue_model import create_espn_venue_model


def _create_espn_team(
    competitor: Dict[str, Any],
    odds_dict: Dict[str, Any],
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    league: League,
) -> TeamModel:
    team_response = session.get(competitor["team"]["$ref"])
    team_response.raise_for_status()
    team_dict = team_response.json()

    odds_key = competitor["homeAway"] + "TeamOdds"
    odds: list[OddsModel] = []
    if odds_dict:
        odds = [  # pyright: ignore
            create_espn_odds_model(
                x[odds_key],
                create_espn_bookie_model(x["provider"]),
            )
            for x in odds_dict["items"]
            if odds_key in x and MONEYLINE_KEY in x[odds_key]
        ]

    roster_dict = {}
    if "roster" in competitor:
        roster_response = session.get(competitor["roster"]["$ref"])
        roster_response.raise_for_status()
        roster_dict = roster_response.json()

    score_response = session.get(competitor["score"]["$ref"])
    score_response.raise_for_status()
    score_dict = score_response.json()

    return create_espn_team_model(
        session,
        team_dict,
        roster_dict,
        odds,
        score_dict,
        dt,
        league,
    )


def _create_venue(
    event: dict[str, Any], session: requests_cache.CachedSession, dt: datetime.datetime
) -> VenueModel | None:
    venue = None
    if "venue" in event:
        venue = create_espn_venue_model(event["venue"], session, dt)
    if venue is None and "venues" in event:
        venues = event["venues"]
        if venues:
            venue_url = event["venues"][0]["$ref"]
            venue_response = session.get(venue_url)
            venue_response.raise_for_status()
            venue = create_espn_venue_model(venue_response.json(), session, dt)
    return venue  # pyright: ignore


def _create_teams(
    event: dict[str, Any],
    session: requests_cache.CachedSession,
    venue: VenueModel | None,
    dt: datetime.datetime,
    league: League,
) -> tuple[list[TeamModel], int | None, datetime.datetime | None]:
    # pylint: disable=too-many-locals
    teams = []
    attendance = None
    end_dt = None
    for competition in event["competitions"]:
        odds_dict = {}
        if "odds" in competition:
            odds_response = session.get(competition["odds"]["$ref"])
            odds_response.raise_for_status()
            odds_dict = odds_response.json()

        for competitor in competition["competitors"]:
            teams.append(_create_espn_team(competitor, odds_dict, session, dt, league))
        attendance = competition["attendance"]
        if "situation" in competition:
            situation_url = competition["situation"]["$ref"]
            situation_response = session.get(situation_url)
            situation_response.raise_for_status()
            situation = situation_response.json()
            if "lastPlay" in situation:
                last_play_response = session.get(situation["lastPlay"]["$ref"])
                last_play_response.raise_for_status()
                last_play = last_play_response.json()
                if "wallclock" in last_play:
                    end_dt = parse(last_play["wallclock"])
        if venue is not None and end_dt is not None:
            end_dt = localize(venue, end_dt)
    return teams, attendance, end_dt


def _create_espn_game_model(
    event: dict[str, Any],
    week: int,
    game_number: int,
    session: requests_cache.CachedSession,
    league: League,
    year: int | None,
    season_type: SeasonType | None,
) -> GameModel:
    dt = parse(event["date"])
    venue = _create_venue(event, session, dt)
    if venue is not None:
        dt = localize(venue, dt)
    teams, attendance, end_dt = _create_teams(event, session, venue, dt, league)
    return GameModel(
        dt=dt,
        week=week,
        game_number=game_number,
        venue=venue,
        teams=teams,
        end_dt=end_dt,
        attendance=attendance,
        league=str(league),
        year=year,
        season_type=season_type,
        postponed=None,
        play_off=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_espn_game_model(
    event: dict[str, Any],
    week: int,
    game_number: int,
    session: requests_cache.CachedSession,
    league: League,
    year: int | None,
    season_type: SeasonType | None,
) -> GameModel:
    return _create_espn_game_model(
        event, week, game_number, session, league, year, season_type
    )


def create_espn_game_model(
    event: dict[str, Any],
    week: int,
    game_number: int,
    session: requests_cache.CachedSession,
    league: League,
    year: int | None,
    season_type: SeasonType | None,
) -> GameModel:
    """Creates an ESPN game model."""
    dt = parse(event["date"])
    if (
        not pytest_is_running.is_running()
        and dt < datetime.datetime.now() - datetime.timedelta(days=7)
    ):
        return _cached_create_espn_game_model(
            event, week, game_number, session, league, year, season_type
        )
    with session.cache_disabled():
        return _create_espn_game_model(
            event, week, game_number, session, league, year, season_type
        )



"""AFL AFL league model."""

# pylint: disable=too-many-statements,protected-access,too-many-locals,bare-except,too-many-branches
import datetime
import logging
import re
from typing import Iterator

import requests_cache
from bs4 import BeautifulSoup
from playwright.sync_api import sync_playwright

from ....playwright import ensure_install
from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from .afl_afl_game_model import create_afl_afl_game_model


def _parse_game_info(
    html: str, session: requests_cache.CachedSession, ladder: list[str]
) -> Iterator[GameModel]:
    soup = BeautifulSoup(html, "lxml")
    for div in soup.find_all("div", {"class": re.compile(".*js-match-list-item.*")}):
        team_names = []
        for span in div.find_all(
            "span", {"class": re.compile(".*team-lineups__team-name.*")}
        ):
            team_names.append(span.get_text().strip())
        teams_players: list[list[tuple[str, str, str, str]]] = [[], []]
        for div_positions_row in div.find_all(
            "div", {"class": re.compile(".*team-lineups__positions-row.*")}
        ):
            row_players: list[list[tuple[str, str, str, str]]] = []
            for div_team_players in div_positions_row.find_all(
                "div",
                {"class": re.compile(".*team-lineups__positions-players.*")},
            ):
                team_players: list[tuple[str, str, str, str]] = []
                for a in div_team_players.find_all(
                    "a", {"class": re.compile(".*js-player-profile-link.*")}
                ):
                    player_id = "afl:" + a.get("data-player-id")
                    first_name = a.get("data-first-name")
                    second_name = a.get("data-surname")
                    player_number = None
                    for span_player_number in a.find_all(
                        "span",
                        {"class": re.compile(".*team-lineups__player-number.*")},
                    ):
                        player_number = (
                            span_player_number.get_text()
                            .strip()
                            .replace("[", "")
                            .replace("]", "")
                        )
                    if player_number is None:
                        raise ValueError("player_number is null")
                    team_players.append(
                        (player_id, player_number, first_name, second_name)
                    )
                row_players.append(team_players)
                if len(row_players) >= 2:
                    break
                for count, row_players_list in enumerate(row_players):
                    try:
                        teams_players[count].extend(row_players_list)
                    except IndexError as exc:
                        # logging.error(html)
                        logging.error(len(team_players))
                        logging.error(count)
                        raise exc
        dt = None
        for time in div.find_all(
            "time", {"class": re.compile(".*match-list-alt__header-time.*")}
        ):
            dt = datetime.datetime.fromtimestamp(float(time.get("data-date")) / 1000.0)
        venue_name: str | None = None
        for span in div.find_all(
            "span", {"class": re.compile(".*match-list-alt__header-venue.*")}
        ):
            venue_name = span.get_text().strip().replace(",", "")
        if venue_name is None:
            raise ValueError("venue_name is null")
        if dt is None:
            raise ValueError("dt is null")
        yield create_afl_afl_game_model(
            team_names, teams_players, dt, venue_name, session, ladder
        )


class AFLAFLLeagueModel(LeagueModel):
    """AFL AFL implementation of the league model."""

    def __init__(
        self, session: requests_cache.CachedSession, position: int | None = None
    ) -> None:
        super().__init__(League.AFL, session)

    @property
    def _ladder(self) -> list[str]:
        ladder = []
        ensure_install()
        with sync_playwright() as p:
            browser = p.chromium.launch()
            context = browser.new_context()
            page = context.new_page()
            try:
                page.goto(
                    "https://www.afl.com.au/ladder",
                    wait_until="networkidle",
                    timeout=60000.0,
                )
            except:  # noqa: E722
                logging.warning("Ladder URL timed out.")
            soup = BeautifulSoup(page.content(), "lxml")
            for span in soup.find_all(
                "span", {"class": re.compile(".*stats-table__club-name.*")}
            ):
                team_name = span.get_text().strip()
                if team_name in ladder:
                    continue
                ladder.append(team_name)
        return ladder

    @property
    def games(self) -> Iterator[GameModel]:
        ladder = self._ladder
        ensure_install()
        with sync_playwright() as p:
            browser = p.chromium.launch()
            context = browser.new_context()
            page = context.new_page()
            page.goto(
                "https://www.afl.com.au/matches/team-lineups", wait_until="networkidle"
            )
            yield from _parse_game_info(page.content(), self.session, ladder)



"""Aussportsbetting team model."""

# pylint: disable=too-many-arguments
import datetime

import requests_cache

from ...cache import MEMORY
from ..google.google_news_model import create_google_news_models
from ..league import League
from ..team_model import TeamModel
from ..x.x_social_model import create_x_social_model
from .aussportsbetting_odds_model import create_aussportsbetting_odds_model


@MEMORY.cache(ignore=["session"])
def create_aussportsbetting_team_model(
    name: str,
    points: float,
    odds: float,
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    league: League,
) -> TeamModel:
    """Create a team model based off aus sports betting."""
    odds_model = create_aussportsbetting_odds_model(odds)
    return TeamModel(
        identifier=name,
        name=name,
        location=None,
        players=[],
        odds=[odds_model],  # pyright: ignore
        points=points,
        ladder_rank=None,
        news=create_google_news_models(name, session, dt, league),
        social=create_x_social_model(name, session, dt),
        field_goals=None,
    )



"""The CLI for executing the data harvesting."""

import io
import logging
import sys

from . import __VERSION__
from .args import parse_args
from .data.league import league_from_str
from .logger import setup_logger
from .sportsball import SportsBall

_STDOUT_FILE = "-"


def main() -> None:
    """The main CLI function."""
    args = parse_args()
    setup_logger()

    logging.info("--- sportsball %s ---", __VERSION__)

    ball = SportsBall()
    league = ball.league(league_from_str(args.league))
    df = league.to_frame()
    handle = io.BytesIO()
    df.to_parquet(handle, compression="gzip")
    handle.seek(0)
    if args.file == _STDOUT_FILE:
        sys.stdout.buffer.write(handle.getbuffer())
    else:
        with open(args.file, "wb") as fhandle:
            fhandle.write(handle.getbuffer())


if __name__ == "__main__":
    main()



"""Aussportsbetting odds model."""

from ...cache import MEMORY
from ..team_model import OddsModel
from .aussportsbetting_bookie_model import create_aussportsbetting_bookie_model


@MEMORY.cache
def create_aussportsbetting_odds_model(odds: float) -> OddsModel:
    """Create an odds model based off aus sports betting."""
    bookie = create_aussportsbetting_bookie_model()
    return OddsModel(odds=odds, bookie=bookie, dt=None)  # pyright: ignore



"""The prototype class for social."""

import datetime

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType


class SocialModel(BaseModel):
    """The serialisable social class."""

    network: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    post: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    comments: int
    reposts: int
    likes: int
    views: int | None
    published: datetime.datetime



"""Combined weather model."""

from ..weather_model import WeatherModel


def create_combined_weather_model(
    weather_models: list[WeatherModel],
) -> WeatherModel | None:
    """Create a weather model by combining many weather models."""
    if not weather_models:
        return None
    temperature = None
    relative_humidity = None
    for weather_model in weather_models:
        weather_model_temperature = weather_model.temperature
        if weather_model_temperature is not None:
            temperature = weather_model_temperature
        weather_model_relative_humidity = weather_model.relative_humidity
        if weather_model_relative_humidity is not None:
            relative_humidity = weather_model_relative_humidity
    return WeatherModel(temperature=temperature, relative_humidity=relative_humidity)
